from rpyc_ws import connect_ws

with connect_ws("ws://localhost:8000/rpyc-ws/") as conn:
    print(conn.modules.os.getcwd())

# or

conn = connect_ws("ws://localhost:8000/rpyc-ws/")
print(conn.modules.os.getcwd())
conn.close()



import asyncio

from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from rpyc.utils.classic import connect_stream

from rpyc_ws.stream import CallbackStream


def create_rpyc_fastapi_app(path: str = "/rpyc-ws/"):
    app = FastAPI()

    @app.websocket(path)
    async def _rpyc_ws(websocket: WebSocket):
        await websocket.accept()
        loop = asyncio.get_running_loop()

        def ws_close():
            asyncio.run_coroutine_threadsafe(websocket.close(), loop).result()

        def ws_receive_bytes(timeout: float):
            future = asyncio.run_coroutine_threadsafe(websocket.receive_bytes(), loop)
            try:
                return future.result(timeout)
            except TimeoutError:
                return None
            except WebSocketDisconnect as exc:
                raise EOFError("WS closed") from exc

        def ws_send_bytes(data: bytes):
            try:
                asyncio.run_coroutine_threadsafe(
                    websocket.send_bytes(data), loop
                ).result()
            except WebSocketDisconnect as exc:
                raise EOFError("WS closed") from exc

        stream = CallbackStream(
            ws_receive_bytes,
            ws_send_bytes,
            ws_close,
        )
        conn = await loop.run_in_executor(None, connect_stream, stream)
        try:
            await loop.run_in_executor(None, conn.serve_all)
        finally:
            await loop.run_in_executor(None, conn.close)

    return app



from typing import Callable

from rpyc.core.consts import STREAM_CHUNK
from rpyc.core.stream import Stream
from rpyc.lib import Timeout


class CallbackStream(Stream):
    """RPyC stream that uses callbacks to read and write to the stream."""

    __slots__ = ("ws_recv", "ws_send", "ws_close", "ws_closed", "_buf")
    MAX_IO_CHUNK = STREAM_CHUNK

    def __init__(
        self,
        ws_recv: Callable[[float], bytes],
        ws_send: Callable[[bytes], None],
        ws_close: Callable[[], None],
    ):
        self.ws_recv = ws_recv
        self.ws_send = ws_send
        self.ws_close = ws_close
        self._buf = bytearray()

    def close(self):
        self._buf.clear()

    @property
    def closed(self) -> bool:
        return False

    def read(self, count: int):
        out = bytearray()

        # Drain buffered spill-over first
        if self._buf:
            take = self._buf[:count]
            out.extend(take)
            del self._buf[: len(take)]

        while len(out) < count:
            msg = self.ws_recv(None)
            if not msg:
                continue

            needed = count - len(out)
            out.extend(msg[:needed])
            if len(msg) > needed:
                self._buf.extend(msg[needed:])

        return bytes(out)

    def write(self, data: bytes | bytearray | memoryview):
        mv = memoryview(data)
        idx = 0
        try:
            while idx < len(mv):
                chunk = mv[idx : idx + self.MAX_IO_CHUNK]
                self.ws_send(chunk.tobytes())
                idx += len(chunk)
        except Exception as exc:
            self.close()
            raise EOFError("connection closed while writing") from exc

    def poll(self, timeout: float | Timeout | None) -> bool:
        timeout = Timeout(timeout)
        timeout = timeout.timeleft()
        if timeout is not None and timeout <= 0:
            return False

        msg = self.ws_recv(timeout)
        if not msg:
            return False

        self._buf.extend(msg)
        return True



from functools import wraps

from rpyc.utils.classic import connect_stream
from websockets import ConnectionClosed
from websockets.sync.client import connect

from rpyc_ws.stream import CallbackStream


@wraps(connect)
def connect_ws(*args, **kwargs):
    websocket = connect(*args, **kwargs)

    def receive_bytes(timeout: float):
        try:
            result = websocket.recv(timeout, False)
            return result
        except TimeoutError:
            return None
        except ConnectionClosed as exc:
            raise EOFError("WS closed") from exc

    def send_bytes(data: bytes):
        try:
            websocket.send(data)
        except ConnectionClosed as exc:
            raise EOFError("WS closed") from exc

    def close():
        websocket.close()

    stream = CallbackStream(receive_bytes, send_bytes, close)
    return connect_stream(stream)
